# -*- coding: utf-8 -*-
try:
    import pygtk
    pygtk.require('2.0')
except ImportError:
    print ("PyGTK not found. Please make sure it is installed properly and referenced in your PYTHONPATH environment variable.")

import logging
import os.path
import threading

import dbus
import gobject
import gtk
from gettext import gettext as _
from gettext import ngettext

import dancer
import game_selector
import model
import view
import printing
import saver
import sudoku_generator_gui
import sudoku_maker
import timer
import tracker_box
from defaults import (APPNAME, APPNAME_SHORT, AUTHORS, COPYRIGHT, DESCRIPTION,
        LICENSE, MIN_NEW_PUZZLES, UI_DIR, VERSION, WEBSITE, WEBSITE_LABEL)
from gtk_goodies import gconf_wrapper, dialog_extras
import undo

import contact_selector
import networking
import tp_tube
import tube_handler

logging.basicConfig()
logger = logging.getLogger("main")


class SudokuGame(gconf_wrapper.GConfWrapper):

    initial_prefs = {'group_size':9,
                     'always_show_hints':False,
                     'difficulty':0.0,
                     'minimum_number_of_new_puzzles':MIN_NEW_PUZZLES,
                     'highlight':False,
                     'bg_color':'black',
                     'show_tracker':False,
                     'width': 700,
                     'height': 675,
                     'auto_save_interval':60 # auto-save interval in seconds...
                     }
    _tube_service = 'org.gnome.Sudoku'
    _view_obj_path = "/SideView"
    _initiator_obj_path = "/Initiator"
    _receiver_obj_path = "/Receiver"

    def __init__ (self):
        gconf_wrapper.GConfWrapper.__init__(self,
                                            gconf_wrapper.GConf('gnome-sudoku')
                                            )
        self.sudoku_maker = sudoku_maker.SudokuMaker()
        self.sudoku_tracker = saver.SudokuTracker()

        self._main_model = None
        self._notes_model = None
        self._main_grid_vew = None
        self._side_grid_vew = None

        self._history_manager = None
        self._tube = None
        self._peer = None
        self._is_initiator = None
        self._is_waiting_for_response = None # this should be extraced out

        self.setup_gui()

        self.tube_handler = tube_handler.TubeHandler(self._tube_service,
                tube_received_cb=self._process_accepted_tube)

        self.timer = timer.ActiveTimer(self.w)
        self.won = False
        self.worker_connections = []
        self.is_fullscreen = False

        self._auto_open_game()

        self.show()
        # generate puzzles while our use is working...
        self._run_background_generator()

    def _auto_open_game(self):
        game = self._auto_load_saved()
        game_type = game_selector.NewOrSavedGameSelector.SAVED_GAME

        if not game:
            game = self._generate_new_puzzle()
            game_type = game_selector.NewOrSavedGameSelector.NEW_GAME

        if game:
            self._open_game(game_type, game)
        else:
            logger.error("can't open game")
            self.quit_cb()

    def _auto_load_saved(self):
        """Automatically open a saved game

        To keep things simple and keep aligned with other game, we should only
        save *one* game in the future. And auto load this one game on launch"""
        saved_games = self.sudoku_tracker.list_saved_games()
        if saved_games:
            return saved_games[0]
        else:
            logger.debug("no saved game found")
            return None

    def _generate_new_puzzle(self, level="easy"):
        # new_puzzle is a list, with the form of (<puzzle>, <difficulty>)
        puzzles = self.sudoku_maker.get_puzzles(1, [level])
        if puzzles:
            puzzle = puzzles[0][0].replace(" ", "") # delete all spaces
            logger.debug("%s game generated: %s" % (level, puzzle))
            return puzzle
        return None

    def _open_game(self, game_type, puzzle):
        """Finally enter the puzzle"""
        self._puzzle = puzzle
        if game_type == game_selector.NewOrSavedGameSelector.NEW_GAME:
            self._main_model = model.SudokuModel(puzzle, 9)
            self._notes_model = model.NotesModel(self._main_model, group_size=9)
            self._main_grid_vew.connect_to_model(self._main_model)
        elif game_type == game_selector.NewOrSavedGameSelector.SAVED_GAME:
            self._open_saved_game(puzzle)
        else:
            logger.fatal("shouldn't be here")

        self._post_open_setup()

    def _open_saved_game(self, jar):
        virgin, in_prog = jar["game"].split('\n')
        self._main_model = model.SudokuModel(virgin, 9, in_prog)
        self._main_model.set_autofill_count(jar["gsd.auto_fills"])

        # Do the connection here; it will setup the readonly appearance.
        # And it must be before the following tracker-setup
        self._main_grid_vew.connect_to_model(self._main_model)

        self._notes_model = model.NotesModel(self._main_model, group_size=9)
        self._notes_model.set_hints_count(jar["gsd.hints"])
        for (x, y, top, bottom) in jar["notes"]:
            self._notes_model.set_notes(x, y, top, bottom)

        self.timer.__absolute_start_time__ = jar["timer.__absolute_start_time__"]
        self.timer.active_time = jar["timer.active_time"]
        self.timer.total_time = jar["timer.total_time"]

        # this is a bit easily breakable... we take advantage of the fact
        # that we create tracker IDs sequentially and that {}.items()
        # sorts by keys by default
        for tracker, tracked in jar.get('trackers', {}).items():
            # add 1 tracker per existing tracker...
            self._tracker_ui.add_tracker()
            for x, y, val in tracked:
                self._main_model.set_value(x, y, val)
        for tracker, tracking in jar.get('tracking', {}).items():
            if tracking:
                self._tracker_ui.select_tracker(tracker)


    def _post_open_setup(self):
        """Properly do the setup after entering game

        Like show highlight, update hints, start timer, etc."""
        self._main_grid_vew.connect_to_notes_model(self._notes_model)
        self._main_model.connect('puzzle-finished', self._puzzle_finished_cb)
        self._calc_difficulty()
        self.setup_toggles()
        self.update_statusbar()

        self.timer.start_timing()

    def _calc_difficulty(self):
        puzzle = self._main_model.grid.virgin.to_string()
        self._difficulty = self.sudoku_maker.get_difficulty(puzzle)

    def _run_background_generator (self):
        """Generate puzzles in background"""
        if self.gconf['generate_puzzles_in_background']:
            gobject.timeout_add_seconds(1,
                    lambda *args: self.start_worker_thread() and True)

    def _select_game(self):
        """Run the game-selector

        Return (chosen-game-type, chosen-puzzle),
        Return None if user cancels."""
        self.update_statusbar() # make the statusbar empty
        selector = game_selector.NewOrSavedGameSelector()
        choice = selector.run_dialog()
        selector.dialog.destroy()
        return choice

    def show (self):
        self.w.show()

    def setup_gui (self):
        self.initialize_prefs()
        self.setup_main_window()
        self.setup_undo()
        self._setup_main_boxes()

        self.setup_color()
        self.setup_actions()
        self.setup_autosave()
        return
#        self.w.add_accel_group(self.uimanager.get_accel_group())

    def initialize_prefs (self):
        for k, v in self.initial_prefs.items():
            try:
                self.gconf[k]
            except:
                self.gconf[k] = v

    def setup_main_window (self):
        gtk.window_set_default_icon_name('gnome-sudoku')
        self.builder = gtk.Builder()
        self.builder.add_from_file(os.path.join(UI_DIR, 'main.ui'))
        self.w = self.builder.get_object("window")
        self.w.set_default_size(self.gconf['width'], self.gconf['height'])
        self.w.set_title(APPNAME_SHORT)
        self.w.connect('configure-event', self.resize_cb)
        self.w.connect('delete-event', self.quit_cb)
        self.statusbar = self.builder.get_object("statusbar")
        self.toolbar = self.builder.get_object("toolbar")

    def _setup_main_boxes(self):
        main_grid_container = self.builder.get_object("main_grid_container")
        side_grid_container = self.builder.get_object("side_grid_container")
        tracker_ui_container = self.builder.get_object("tracker_ui_container")

        self._main_grid_vew = view.SudokuView(9, self._history_manager)
        main_grid_container.pack_start(self._main_grid_vew, padding=6)

        self._tracker_ui = tracker_box.TrackerBox(self._main_grid_vew)
        self._tracker_ui.hide()
        tracker_ui_container.pack_start(self._tracker_ui)

        self._side_grid_vew = view.SudokuSideView(group_size=9)
        self._side_grid_vew.hide()
        side_grid_container.add(self._side_grid_vew)


    def setup_color (self):
        # setup background colors
        bgcol = self.gconf['bg_color']
        if bgcol != '':
            self._main_grid_vew.set_bg_color(bgcol)

    def setup_actions (self):
        toggle_actions = [
                ("AlwaysShowPossible", self.auto_hint_cb),
                ("ShowImpossibleImplications", self.impossible_implication_cb),
                ("ToggleTracker", self.tracker_toggle_cb),
                ("ToggleToolbar", self.toggle_toolbar_cb),
                ("ToggleHighlight", self.toggle_highlight_cb),
                ("BackgroundGenerator", self.toggle_generator_cb)]
        for action_name, callback in toggle_actions:
            action = self.builder.get_object(action_name)
            action.connect("toggled", callback)

        actions = [
                ("New", self.new_cb),
                ("NewWithContact", self._new_with_contact_cb),
                ("Print", self.print_game),
                ("PrintMany", self.print_multiple_games),
                ("Close", self.quit_cb),
                ("ShowPossible", self.show_hint_cb),
                ("AutofillCurrentSquare", self.auto_fill_current_square_cb),
                ("Autofill", self.auto_fill_cb),
                ("FullScreen", self.full_screen_cb),
                ("Generator", self.generate_puzzle_gui),
                ("PuzzleInfo", self.show_info_cb),
                ("About", self.show_about),
                ("ShowHelp", self.show_help)]
        for action_name, callback in actions:
            action = self.builder.get_object(action_name)
            action.connect("activate", callback)

        edit_actions = [
                ("Clear", self.clear_cb),
                ("ClearNotes", self.clear_notes_cb),
                ]
        for action_name, callback in edit_actions:
            action = self.builder.get_object(action_name)
            action.connect("activate", callback)

    def setup_undo (self):
        undo_action = self.builder.get_object('Undo')
        redo_action = self.builder.get_object('Redo')
        self._history_manager = undo.HistoryManager(undo_action, redo_action)
        undo_action.connect("activate",
                lambda *args: self._history_manager.undo())
        redo_action.connect("activate",
                lambda *args: self._history_manager.redo())

    def setup_autosave (self):
        gobject.timeout_add_seconds(self.gconf['auto_save_interval'] or 60, # in seconds...
                            self.autosave)

    def setup_toggles (self):
        logger.debug("sync up toggles with gconf values")
        [self.gconf_wrap_toggle(conf, action) for (conf, action) in [
            ('always_show_hints',
                self.builder.get_object('AlwaysShowPossible')),
            ('show_impossible_implications',
                self.builder.get_object('ShowImpossibleImplications')),
            ('generate_puzzles_in_background',
                self.builder.get_object('BackgroundGenerator')),
            ('show_toolbar',
                self.builder.get_object('ToggleToolbar')),
            ('highlight',
                self.builder.get_object('ToggleHighlight')),
            ('show_tracker',
                self.builder.get_object('ToggleTracker'))]]

    def start_worker_thread(self):
        n_new_puzzles = self.sudoku_maker.n_puzzles(new = True)
        if n_new_puzzles < self.gconf['minimum_number_of_new_puzzles']:
            self.worker = threading.Thread(target = lambda *args: self.sudoku_maker.work(limit = 5))
            self.worker_connections = [
                self.timer.connect('timing-started', self.sudoku_maker.resume),
                self.timer.connect('timing-stopped', self.sudoku_maker.pause)
                ]
            self.worker.start()
        return True

    def stop_worker_thread(self):
        if hasattr(self, 'worker'):
            self.sudoku_maker.stop()
            for c in self.worker_connections:
                self.timer.disconnect(c)

    def stop_dancer(self):
        if hasattr(self, 'dancer'):
            self.dancer.stop_dancing()
            delattr(self, 'dancer')

    def _build_report_label(self):
        sublabel = _("You completed the puzzle in %(totalTime)s (%(activeTime)s active)") % {'totalTime': self.timer.total_time_string(),
        'activeTime': self.timer.active_time_string()
                }
        sublabel += "\n"
        sublabel += ngettext("You got %(n)s hint", "You got %(n)s hints",
                self._notes_model.get_hints_count()) % {
                        'n': self._notes_model.get_hints_count()}
        sublabel += "\n"
        impossible_hints = self._main_model.get_impossible_hints()
        if impossible_hints:
            sublabel += ngettext("You had %(n)s impossibility pointed out.",
                                 "You had %(n)s impossibilities pointed out.",
                                 impossible_hints) % {
                                         'n': impossible_hints}
            sublabel += "\n"
        autofill_count = self._main_model.get_autofill_count()
        if autofill_count:
            sublabel += ngettext("You used the auto-fill %(n)s time",
                                 "You used the auto-fill %(n)s times",
                                 autofill_count) % {
                                         'n': autofill_count}

        return sublabel

    def _puzzle_finished_cb(self, sudoku_model):
        if hasattr(self, 'dancer'):
            return
        self.won = True
        # increase difficulty for next time.
        self.gconf['difficulty'] = self.gconf['difficulty'] + 0.1
        self.timer.finish_timing()
        self.sudoku_tracker.finish_game(self._jar_game())
        self.dancer = dancer.GridDancer(self._main_grid_vew, self._main_model)
        self.dancer.start_dancing()

        if not self._tube: # local game
            label = self._build_report_label()
        else:
            if self._negotiate_result():
                label = "You WIN"
                self._stop_network_game()
            else:
                return

        dialog_extras.show_message(_("You win!"), label = _("You win!"),
                                   sublabel=label)

    def _negotiate_result(self):
        def reply_handler(*args):
            logger.debug('signal peer: %s' % args)
        def error_handler(error):
            logger.error('signal peer: %s' % error)

        i_win = None
        if self._is_initiator:
            i_win = True
            logger.debug('signal receiver that they fail')
            self._peer.acknowledge_win(
                    reply_handler=reply_handler,
                    error_handler=error_handler)
        else:
            assert self._peer is not None
            if not self._peer.has_won(): # FIXME should use async all
                i_win = True
                logger.debug('signal initiator that they fail')
                self._peer.acknowledge_win(
                        reply_handler=reply_handler,
                        error_handler=error_handler)
            # else do nothing; wait for initatior to tell us that we fail
        return i_win

    def fail_network_game(self):
        """We fail the game
        """
        dialog = gtk.Dialog("You fail",
                self.w,
                gtk.DIALOG_MODAL | gtk.DIALOG_DESTROY_WITH_PARENT,
                (gtk.STOCK_OK, gtk.RESPONSE_OK))
        label = gtk.Label("you fail")
        label.show()
        dialog.vbox.pack_start(label)
        dialog.run()
        dialog.destroy()

    def new_cb (self, *args):
        if self._close_current_game():
            self._auto_open_game()
            return

    def _new_with_contact_cb(self, action):
        def response_cb(dialog, response_id):
            if self._is_waiting_for_response:
                if response_id in (gtk.RESPONSE_CANCEL,
                        gtk.RESPONSE_DELETE_EVENT):
                    self._tube_offer.cancel()
                    self._is_initiator = None
            # else don't need to do anything
            self._wait_dialog.destroy()

        def offer_fail_cb(offer, reason):
            self._is_waiting_for_response = False
            self._is_initiator = None
            notebook = builder.get_object("notebook")
            notebook.set_current_page(1)
            button_cancel = builder.get_object("button_cancel")
            button_ok = builder.get_object("button_ok")
            button_cancel.hide()
            button_ok.show()

        choice = contact_selector.contact_selector_run(self.w)
        if not choice:
            return

        self._is_initiator = True
        conn, handle, mode, level = choice

        puzzle = self._generate_new_puzzle(level)
        if not puzzle:
            logger.error("failed to genereate new puzzle")
            # should popup?
            return

        self._tube_offer = tp_tube.TubeOffer(
                conn, handle, self._tube_service, puzzle, mode)

        builder = gtk.Builder()
        builder.add_from_file(os.path.join(UI_DIR, "offering_tube.ui"))
        self._wait_dialog = builder.get_object("dialog")

        self._wait_dialog.set_transient_for(self.w)
        self._wait_dialog.connect("response", response_cb)

        self._tube_offer.connect("offer-succeeded",
                self._tube_offered_cb, puzzle, mode)
        self._tube_offer.connect("offer-failed", offer_fail_cb)

        self._tube_offer.execute()
        self._is_waiting_for_response = True
        self._wait_dialog.run()

    def _process_accepted_tube(self, bus, tube, puzzle, mode):
        """User has accepted the tube
        """

        self._is_initiator = False
        self._tube = tube
        self._tube[tp_tube.CHANNEL_INTERFACE].connect_to_signal(
                'Closed', self._tube_closed_cb)

        if not self._negotiate_peer_proxy(bus):
            return

        self.do_stop()
        self._open_game(game_selector.NewOrSavedGameSelector.NEW_GAME, puzzle)
        self._init_side_view_display(puzzle)
        self._negotiate_sideview_proxy(bus, mode)

    def _tube_offered_cb(self, offer, bus, tube, puzzle, mode):
        """Called when the peer accepts our tube"""
        self._wait_dialog.destroy()
        self.do_stop()
        self._tube = tube

        self._tube[tp_tube.CHANNEL_INTERFACE].connect_to_signal(
                'Closed', self._tube_closed_cb)
        self._open_game(game_selector.NewOrSavedGameSelector.NEW_GAME, puzzle)

        if not self._negotiate_peer_proxy(bus):
            return

        self._init_side_view_display(puzzle)
        self._negotiate_sideview_proxy(bus, mode)

    def _negotiate_peer_proxy(self, bus):
        if self._is_initiator:
            me, peer = self._initiator_obj_path, self._receiver_obj_path
        else:
            me, peer = self._receiver_obj_path, self._initiator_obj_path

        myself = networking.Peer(bus, me, self)
        try:
            self._peer = bus.get_object(object_path=peer)
        except dbus.exceptions.DBusException as err:
            logger.error("Error when contacting with initiator: %s" % err)
            return False
        return True

    def get_puzzle(self):
        return self._puzzle.replace(" ", "")

    def _init_side_view_display(self, puzzle):
        for x in range(9):
            for y in range(9):
                val = int(puzzle[9 * y + x])
                if val:
                    self._side_grid_vew.set_value(x, y, val)

    def _negotiate_sideview_proxy(self, bus, mode):
        receiver_view = self._view_obj_path + "/receiver"
        initiator_view = self._view_obj_path + "/initiator"
        if self._is_initiator:
            our_obj_path, peer_obj_path = initiator_view, receiver_view
        else:
            our_obj_path, peer_obj_path = receiver_view, initiator_view

        # First export our own side-grid-view on dbus
        proxy_of_side_view = networking.SideViewProxy(bus, our_obj_path, mode,
                 self._side_grid_vew)
        logger.debug("exported side-view on dbus")
        self._side_grid_vew.show()

        # Then fetch our peer's side-grid-view
        _remote_view_proxy = bus.get_object(object_path=peer_obj_path)
        logger.debug("got a remote view which we can manipulate")
        # we issue command on remote_view_handler, which will communicate with
        # remote_view
        view_handler = networking.RemoteViewHandler(_remote_view_proxy)
        # FIXME this is ugly, but we need a way to disconnect this callback
        # when we are disconnected
        self._view_handler_cb = self._main_grid_vew.connect("focus-changed",
                lambda view, x, y: view_handler.set_focus(x, y))
        self._main_model.add_observer(view_handler)

    def _close_current_game(self):
        """Close current running game

        Return True if game is closed, or else return False"""
        if self._tube: # network game:
            self._stop_network_game()

        if self._should_save():
            try:
                if dialog_extras.getBoolean(
                        label = _("Save current game?"),
                        custom_yes = _("_Save"),
                        custom_no = _("_Abandon")):
                    self._save_game()
                else:
                    self.sudoku_tracker.abandon_game(self._jar_game())
            except dialog_extras.UserCancelledError:
                # User cancelled; current game is not closed
                return False
        self.do_stop()
        return True # also return True if game is not changed at all

    def do_stop (self):
        self.stop_dancer()
        self._main_model = None
        self._notes_model = None
        self._main_grid_vew.reset()
        self._tracker_ui.reset()
        self._side_grid_vew.reset()
        self._history_manager.clear()
        self.won = False

    def resize_cb (self, widget, event):
        self.gconf['width'] = event.width
        self.gconf['height'] = event.height

    def _tube_closed_cb(self):
        self._is_initiator = None
        self._side_grid_vew.hide()
        self._tube = None
        self._main_grid_vew.disconnect(self._view_handler_cb)
        logger.debug("tube has been closed")

    def _stop_network_game(self, *args):
        logger.debug("close tube-channel")
        try:
            self._tube[tp_tube.CHANNEL_INTERFACE].Close()
        except dbus.exceptions.DBusException as err:
            logger.error("can't close tube: %s" % err)
        self._side_grid_vew.reset()
        self._side_grid_vew.hide()
        self._tube = None

    def quit_cb (self, *args):
        self.w.hide()
        if self._tube:
            self._stop_network_game()

        if self._should_save():
            self._save_game()
        # make sure we really go away before doing our saving --
        # otherwise we appear sluggish.
        while gtk.events_pending():
            gtk.main_iteration()
        self.stop_worker_thread()
        # allow KeyboardInterrupts, which calls quit_cb outside the main loop
        try:
            gtk.main_quit()
        except RuntimeError:
            pass

    def _save_game(self):
        self.sudoku_tracker.save_game(self._jar_game())

    def _jar_game(self):
        jar = {} # what we will pickle
        self.timer.finish_timing()
        jar["game"] = self._main_model.grid.to_string()
        jar["trackers"] = self._main_grid_vew.get_trackers()
        jar["tracking"] = self._main_grid_vew.get_trackers_tracking()
        jar["notes"] = self._notes_model.get_all_notes()
        jar["gsd.hints"] = self._notes_model.get_hints_count()
        jar["gsd.auto_fills"] = self._main_model.get_autofill_count()
        jar["timer.__absolute_start_time__"] = \
                self.timer.get_absolute_start_time()
        jar["timer.active_time"] = self.timer.get_active_time()
        jar["timer.total_time"] = self.timer.get_total_time()

        return jar

    def full_screen_cb(self, action):
        if self.is_fullscreen:
            self.w.unfullscreen()
            self.is_fullscreen = False
        else:
            self.w.fullscreen()
            self.is_fullscreen = True

    def clear_cb(self, action):
        self._main_grid_vew.do_clear()

    def clear_notes_cb(self, action):
        self._notes_model.clear_notes()

    def show_hint_cb(self, action):
        current = self._main_grid_vew.get_focus()
        if current:
            self._notes_model.show_hint(current.x, current.y)

    def auto_hint_cb(self, action):
        self._notes_model.toggle_auto_hint(action.get_active())

    def impossible_implication_cb(self, action):
        self._main_model.toggle_impossible_implications(True)

    def auto_fill_cb(self, action):
        cmd = model.AutoFillCmd(self._main_model)
        self._history_manager.execute_command(cmd)

    def auto_fill_current_square_cb(self, action):
        current = self._main_grid_vew.get_focus()
        if current:
            x, y = current.x, current.y
            new = self._main_model.calc_solution(x, y)
            if new:
                old = self._main_grid_vew.get_value(x, y)
                cmd = model.ChangeValueCmd(self._main_model, x, y, old, new)
                self._history_manager.execute_command(cmd)

    def tracker_toggle_cb (self, widg):
        if widg.get_active():
            self._tracker_ui.show()
        else:
            self._tracker_ui.hide()

    def toggle_toolbar_cb (self, widg):
        if widg.get_active():
            self.toolbar.show()
        else:
            self.toolbar.hide()

    def set_statusbar_value (self, status):
        if not hasattr(self, 'sbid'):
            self.sbid = self.statusbar.get_context_id('game_info')
        self.statusbar.pop(self.sbid)
        self.statusbar.push(self.sbid, status)

    def update_statusbar(self):
        if not self._main_model:
            self.set_statusbar_value(" ")
            return True

        puzzle_diff = self._difficulty
        tot_string = _("Playing %(difficulty)s puzzle.") % {
                'difficulty': puzzle_diff.value_string()}
        tot_string += " " + "(%1.2f)" % puzzle_diff.value

        self.set_statusbar_value(tot_string)
        return True

    def toggle_highlight_cb (self, widg):
        self._main_grid_vew.toggle_highlight(widg.get_active())

    def show_info_cb(self, action):
        if not self._main_model.grid:
            dialog_extras.show_message(parent = self.w,
                                       title = _("Puzzle Information"),
                                       label = _("There is no current puzzle.")
                                       )
            return
        information = self._difficulty.to_string()
        dialog_extras.show_message(parent = self.w,
                                   title = _("Puzzle Statistics"),
                                   label = _("Puzzle Statistics"),
                                   sublabel = information)

    def toggle_generator_cb (self, toggle):
        if toggle.get_active():
            self.start_worker_thread()
        else:
            self.stop_worker_thread()

    def autosave (self):
        # this is called on a regular loop and will autosave if we
        # have reason to...
        if self._should_save():
            self._save_game()
        return True

    def _should_save(self):
        return self._main_model and self._main_model.grid.is_changed() and \
                not self.won

    def show_about(self, action):
        about = gtk.AboutDialog()
        about.set_transient_for(self.w)
        about.set_name(APPNAME)
        about.set_version(VERSION)
        about.set_copyright(COPYRIGHT)
        about.set_license(LICENSE[0] + '\n\n' + LICENSE[1] + '\n\n'  + LICENSE[2])
        about.set_wrap_license(True)
        about.set_comments(DESCRIPTION)
        about.set_authors(AUTHORS)
        about.set_website(WEBSITE)
        about.set_website_label(WEBSITE_LABEL)
        about.set_logo_icon_name("gnome-sudoku")
        about.set_translator_credits(_("translator-credits"))
        about.connect("response", lambda d, r: d.destroy())
        about.show()

    def show_help(self, action):
        try:
            gtk.show_uri(self.w.get_screen(), "ghelp:gnome-sudoku", gtk.get_current_event_time())
        except gobject.GError, error:
            # FIXME: This should create a pop-up dialog
            print _('Unable to display help: %s') % str(error)

    def print_game(self, action):
        printing.print_sudokus([self._main_model], self.w)

    def print_multiple_games(self, action):
        gp = printing.GamePrinter(self.sudoku_maker, self.gconf)
        gp.run_dialog()

    def generate_puzzle_gui(self, action):
        sudoku_generator_gui.GameGenerator(self, self.gconf)


def start_game(debug=False):
    if debug:
        print 'Starting GNOME Sudoku in debug mode'
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)

    ##  You must call g_thread_init() before executing any other GLib
    ##  functions in a threaded GLib program.
    gobject.threads_init()

    game = SudokuGame()
    try:
        gtk.main()
    except KeyboardInterrupt:
        # properly quit on a keyboard interrupt...
        game.quit_cb()
