# -*- coding: utf-8 -*-

import gobject

import sudoku


class ParallelDict (dict):
    """A handy new sort of dictionary for tracking conflicts.

    pd = ParallelDict()
    pd[1] = [2, 3, 4] # 1 is linked with 2, 3 and 4
    pd -> {1:[2, 3, 4], 2:[1], 3:[1], 4:[1]}
    pd[2] = [1, 3, 4] # 2 is linked with 3 and 4 as well as 1
    pd -> {1: [2, 3, 4], 2:[3, 4], 3:[1, 2], 4:[1, 2]}
    Now for the cool part...
    del pd[1]
    pd -> {2: [2, 3], 3:[2], 4:[2]}

    Pretty neat, no?
    """
    def __init__ (self, *args):
        dict.__init__(self, *args)

    def __setitem__ (self, k, v):
        dict.__setitem__(self, k, set(v))
        for i in v:
            if i == k:
                continue
            if self.has_key(i):
                self[i].add(k)
            else:
                dict.__setitem__(self, i, set([k]))

    def __delitem__ (self, k):
        v = self[k]
        dict.__delitem__(self, k)
        for i in v:
            if i == k:
                continue
            if self.has_key(i):
                # Make sure we have a reference to i. If we don't
                # something has gone wrong... but according to bug
                # 385937 this has gone wrong at least once, so we'd
                # better check for it.
                if k in self[i]:
                    self[i].remove(k)
                if not self[i]:
                    # If k was the last value in the list of values
                    # for i, then we delete i from our dictionary
                    dict.__delitem__(self, i)


class NumberBoxModel:
    """Used to notify observers for the change of SudokuModel"""
    def __init__(self, x, y, value=None, conflict=None):
        self.x = x
        self.y = y
        self.value = value
        self.conflict = conflict


class NotesModel:
    """Track notes for a SudokuModel

    Act as a observer of SudokuModel, and a model/observable for SudokuView.
    SudokuModel should know nothing about notes.
    """

    def __init__(self, sudoku_model, group_size):
        self._notes = {}
        self._observers = []
        self._group_size = group_size
        self._model = sudoku_model
        self._model.add_observer(self)

        self._always_show_hints = False
        self._hints_count = 0

    def toggle_auto_hint(self, flag):
        if flag:
            self._always_show_hints = True
            self.update_all_hints()
        else:
            self._always_show_hints = False
            self._clear_hints()

    def add_observer(self, observer):
        self._observers.append(observer)
        init_change = []
        for (x, y) in self._notes:
            top, bottom = self._notes[(x, y)]
            init_change.append((x, y, top, bottom))
        self._signal_observers(init_change)

    def _signal_observers(self, changes):
        for observer in self._observers:
            observer.update_notes(changes)

    def set_notes(self, x, y, top_note=None, bottom_note=None):
        if (x, y) not in self._notes:
            self._notes[(x, y)] = [None, None]

        if top_note is not None:
            self._notes[(x, y)][0] = top_note
        if bottom_note is not None:
            self._notes[(x, y)][1] = bottom_note

        change = (x, y, top_note, bottom_note)
        self._signal_observers([change])

    def get_all_notes(self):
        # These are all for keeping compatible with old saving format:
        # notes was saved as a list of (x, y, top, bottom)
        result = []
        for (x, y) in self._notes:
            top, bottom = self._notes[(x, y)]
            result.append((x, y, top, bottom))
        return result

    def update(self, changes):
        if not self._always_show_hints:
            return

        self.update_all_hints()

    def update_all_hints(self):
        for x in range(self._group_size):
            for y in range(self._group_size):
                self._update_hint_for_entry(x, y)

    def show_hint(self, x, y):
        self._update_hint_for_entry(x, y)

    def _update_hint_for_entry(self, x, y):
        # no need to show anything if it's already filled, being correct or not
        if self._model.get_value(x, y):
            self.set_notes(x, y, bottom_note="")
            return

        vals = self._model.grid.possible_values(x, y)
        vals = list(vals)
        vals.sort()
        if vals:
            txt = ''.join([str(v) for v in vals])
            self.set_notes(x, y, bottom_note=txt)
            self._increase_hints_count()
        else:
            self.set_notes(x, y, bottom_note="X")

    def _clear_hints(self):
        self._clear_bottom_note()

    def _clear_top_note(self):
        for x in range(self._group_size):
            for y in range(self._group_size):
                self.set_notes(x, y, top_note="")

    def _clear_bottom_note(self):
        for x in range(self._group_size):
            for y in range(self._group_size):
                self.set_notes(x, y, bottom_note="")

    def _increase_hints_count(self):
        self._hints_count += 1

    def set_hints_count(self, count):
        self._hints_count = count

    def get_hints_count(self):
        return self._hints_count

    def clear_notes(self):
        self._clear_top_note()
        self._clear_bottom_note()

class SudokuModel(gobject.GObject):

    __gsignals__ = {
            # atm. only used by dancer
            "puzzle-finished": (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, ())
            }

    def __init__ (self, virgin_grid, group_size, initial_grid=None):
        gobject.GObject.__init__(self)
        self.autofill_count = 0
        self.show_impossible_implications = False
        self.impossibilities = set()
        self.__trackers_tracking__ = {}
        self.group_size = group_size
        self.grid = None
        self._virgin_grid = None
        self._observers = []

        self.__error_pairs__ = ParallelDict()
        self._setup_virgin_grid(virgin_grid)
        if initial_grid:
            self._setup_initial_grid(initial_grid)

    def add_observer(self, observer):
        self._observers.append(observer)

    def _signal_observers(self, values):
        for observer in self._observers:
            observer.update(values)
        self._check_for_completeness()

    def _setup_virgin_grid(self, virgin):
        """both grid and initial_grid should be str"""
        virgin = virgin.replace(" ", "")
        assert len(virgin) == self.group_size ** 2
        self._virgin_grid = [int(c) for c in virgin]
        self.grid = sudoku.InteractiveSudoku(virgin, group_size=self.group_size)

    def _setup_initial_grid(self, initial_grid):
        values = [int(c) for c in initial_grid.split()]
        for row in range(self.group_size):
            for col in range(self.group_size):
                index = row * 9 + col
                if values[index] and not self.grid._get_(col, row):
                    self.set_value(col, row, values[index])

    def set_value(self, x, y, value):
        old_conflicts = self._remove_old_conficts(x, y)
        new_conflicts = []

        # Some value won't go into the grid, e.g. conflicts values?
        # This makes things complicated.
        if not value:
            if self.grid._get_(x, y):
                self.grid.remove(x, y)
        else:
            try:
                self.grid.add(x, y, value, force=True)
            except sudoku.ConflictError:
                new_conflicts = self._find_new_conflicts(x, y, value)

        flag = True if new_conflicts else False
        result = [NumberBoxModel(x, y, value=value, conflict=flag)]
        result.extend(old_conflicts)
        result.extend(new_conflicts)
        result.extend(self._update_impossible_implications(x, y))

        self._signal_observers(result)

    def get_value(self, x, y):
        return self.grid._get_(x, y)

    def get_virgin_value(self, x, y):
        return self._virgin_grid[9 * y + x]

    def _remove_old_conficts(self, x, y):
        result = []
        old_conflicts = self._remove_related_conflicts(x, y)
        for entry in old_conflicts:
            change = NumberBoxModel(entry[0], entry[1], conflict=False)
            result.append(change)
        return result

    def _find_new_conflicts(self, x, y, value):
        result = []
        new_conflicts = self.grid.find_conflicts(x, y, value)
        if new_conflicts:
            self.__error_pairs__[(x, y)] = new_conflicts
        for entry in new_conflicts:
            change = NumberBoxModel(entry[0], entry[1], conflict=True)
            result.append(change)
        return result

    def _remove_related_conflicts(self, x, y):
        result = []
        if self.__error_pairs__.has_key((x, y)):
            errors_removed = self.__error_pairs__[(x, y)]
            del self.__error_pairs__[(x, y)]
            for coord in errors_removed:
                # If we're not an error by some other pairing...
                if not self.__error_pairs__.has_key(coord):
                    result.append(coord)
        return result

    def _check_for_completeness(self):
        if self.grid.check_for_completeness():
            self.emit("puzzle-finished")

    def set_autofill_count(self, count):
        self.autofill_count = count

    def get_autofill_count(self):
        return self.autofill_count

    def auto_fill (self):
        self.autofill_count += 1
        result = []

        changed = set(self.grid.auto_fill()) # there are duplicate?
        for coords, val in changed:
            x, y = coords
            result.append(NumberBoxModel(x, y, value=val))

        self._signal_observers(result)
        return changed

    def calc_solution(self, x, y):
        return self.grid.calc_solution(x, y)

    def toggle_impossible_implications(self, show):
        self.show_impossible_implications = show
        if not show:
            self._clean_impossible_implications()

    def _clean_impossible_implications(self):
        for (x, y) in self.impossibilities:
            self.set_value(x, y, 0)

    def _update_impossible_implications(self, x, y):
        if not self.show_impossible_implications:
            return []

        changes = []
        new_imposs = set(self.grid.find_impossible_implications(x, y))

        for coord in new_imposs - self.impossibilities:
            # -1 means impossible
            changes.append(NumberBoxModel(coord[0], coord[1], value=-1))
        for coord in self.impossibilities - new_imposs:
            # remove the 'X' mark
            changes.append(NumberBoxModel(coord[0], coord[1], value=0))

        self.impossibilities = new_imposs
        return changes

    def get_impossible_hints(self):
        return len(self.impossibilities)

#
# Commands that can undo themselves
#

class ChangeValueCmd:
    def __init__(self, model, x, y, old, new):
        self.execute = lambda: model.set_value(x, y, new)
        self.undo = lambda: model.set_value(x, y, old)

class MultiChangeValueCmd:
    def __init__(self, model, change_list):
        """change_list should be list of (x, y, old, new)
        """
        self._model = model
        self._list = change_list

    def execute(self):
        for x, y, old, new in self._list:
            self._model.set_value(x, y, new)

    def undo(self):
        for x, y, old, new in self._list:
            self._model.set_value(x, y, old)

class AutoFillCmd:
    def __init__(self, model):
        self._model = model
        self._filled = None

    def execute(self):
        self._filled = self._model.auto_fill()

    def undo(self):
        for coord, value in self._filled:
            # Auto-fill should only fill in black boxes, so its undo action
            # should fill them in 0.
            # But, conflicting values don't go into the grid :( So we can't
            # go back for a conflicting box.
            self._model.set_value(coord[0], coord[1], 0)

