# -*- coding: utf-8 -*-

import os.path

import gtk
import gobject
from gettext import gettext as _

import defaults

STOCK_PIXBUFS = {}

def _add_tracker_stock_icon():
    icon_factory = gtk.IconFactory()
    for filename, stock_id in [('footprints.png', 'tracks'), ]:
        try:
            pixbuf = gtk.gdk.pixbuf_new_from_file(os.path.join(defaults.IMAGE_DIR,
                filename))
        except gobject.GError, err:
            print 'Failed to load pixbuf: %s' % err
            continue
        STOCK_PIXBUFS[stock_id] = pixbuf
        iconset = gtk.IconSet(pixbuf)
        icon_factory.add(stock_id, iconset)
        icon_factory.add_default()

    gtk.stock_add([('tracks',
                    _('Track moves'),
                    0, 0, ""), ])

_add_tracker_stock_icon()

def _pixbuf_transform_color(pixbuf, color):
    """Return new pixbuf with color changed to color"""
    pixbuf_str = pixbuf.get_pixels()
    pixbuf_str_new = ""

    for alpha in pixbuf_str[3::4]:
        pixbuf_str_new += chr(int(color[0]*255))
        pixbuf_str_new += chr(int(color[1]*255))
        pixbuf_str_new += chr(int(color[2]*255))
        pixbuf_str_new += alpha

    return gtk.gdk.pixbuf_new_from_data(pixbuf_str_new, gtk.gdk.COLORSPACE_RGB,
            True, 8, pixbuf.get_width(), pixbuf.get_height(),
            pixbuf.get_rowstride())


class TrackerBox(gtk.VBox):

    def __init__(self, sudoku_view):
        gtk.VBox.__init__(self)
        self.builder = gtk.Builder()
        self.builder.add_from_file(os.path.join(defaults.UI_DIR, 'tracker.ui'))
        self._view_controller = sudoku_view.get_tracker_controller()
        vbox = self.builder.get_object('vbox1')
        vbox.unparent()
        self.pack_start(vbox, expand = True, fill = True)

        self.tracker_tree = None
        self.tracker_model = None
        self.tracker_actions = None

        self._setup_actions()
        self._setup_tree()
        self.show_all()

    def reset (self):
        for tree in self.tracker_model:
            if tree[0] > -1:
                self.tracker_model.remove(tree.iter)

    def _setup_tree (self):
        self.tracker_tree = self.builder.get_object('treeview1')
        self.tracker_model = gtk.ListStore(int, gtk.gdk.Pixbuf, str)
        self.tracker_tree.set_model(self.tracker_model)
        col1 = gtk.TreeViewColumn("", gtk.CellRendererPixbuf(), pixbuf = 1)
        col2 = gtk.TreeViewColumn("", gtk.CellRendererText(), text = 2)
        self.tracker_tree.append_column(col2)
        self.tracker_tree.append_column(col1)
        # Our initial row...
        self.tracker_model.append([-1, None, _('No Tracker')])
        self.tracker_tree.get_selection().connect('changed', self._selection_changed_cb)

    def _setup_actions (self):
        self.tracker_actions = gtk.ActionGroup('tracker_actions')
        self.tracker_actions.add_actions(
            [('Clear',
              gtk.STOCK_CLEAR,
              _('_Clear Tracker'),
              None, _('Clear all moves tracked by selected tracker.'),
              self._clear_cb
              ),
             ('Keep', None,
              _('_Clear Others'),
              None,
              _('Clear all moves not tracked by selected tracker.'),
              self._keep_cb),
             ]
            )
        for action_name, widget_name in [
                ('Clear', 'ClearTrackerButton'),
                ('Keep', 'KeepTrackerButton')]:
            action = self.tracker_actions.get_action(action_name)
            action.connect_proxy(self.builder.get_object(widget_name))
        self.builder.get_object('AddTrackerButton').connect('clicked',
                lambda button: self.add_tracker())
        # Default to insensitive (they only become sensitive once a tracker is added)
        self.tracker_actions.set_sensitive(False)

    def add_tracker (self):
        tracker_id = self._view_controller.create_tracker()
        color = self._view_controller.get_tracker_color(tracker_id)
        pixbuf = _pixbuf_transform_color(STOCK_PIXBUFS['tracks'], color)
        # select our new tracker
        self.tracker_tree.get_selection().select_iter(
            self.tracker_model.append([tracker_id,
                                  pixbuf,
                                  _("Tracker %s") % (tracker_id + 1)]
                                  )
            )

    def select_tracker (self, tracker_id):
        for row in self.tracker_model:
            if row[0] == tracker_id:
                self.tracker_tree.get_selection().select_iter(row.iter)

    def _selection_changed_cb (self, selection):
        mod, itr = selection.get_selected()
        if itr:
            selected_tracker_id = mod.get_value(itr, 0)
        else:
            selected_tracker_id = -1
        # This should be cheap since we don't expect many trackers...
        # We cycle through each row and toggle it off if it's not
        # selected; on if it is selected
        for row in self.tracker_model:
            tid = row[0]
            if tid != -1: # -1 == no tracker
                self._view_controller.toggle_tracker(tid, tid==selected_tracker_id)
        self.tracker_actions.set_sensitive(selected_tracker_id != -1)

    def _clear_cb (self, action):
        mod, itr = self.tracker_tree.get_selection().get_selected()
        # This should only be called if there is an itr, but we'll
        # double-check just in case.
        if itr:
            selected_tracker_id = mod.get_value(itr, 0)
            self._view_controller.delete_by_tracker(selected_tracker_id)

    def _keep_cb (self, action):
        mod, itr = self.tracker_tree.get_selection().get_selected()
        selected_tracker_id = mod.get_value(itr, 0)
        self._view_controller.delete_except_for_tracker(selected_tracker_id)
