# -*- coding: utf-8 -*-

import logging
import os

import dbus
import dbus.service
import gtk
from dbus import PROPERTIES_IFACE
from dbus.mainloop.glib import DBusGMainLoop
from gettext import gettext as _

from telepathy.client.conn import Connection
from telepathy.client.channel import Channel
from telepathy.constants import (
        TUBE_STATE_OPEN,
        SOCKET_ACCESS_CONTROL_CREDENTIALS)
from telepathy.interfaces import (
        CHANNEL_INTERFACE,
        CHANNEL_INTERFACE_TUBE,
        CHANNEL_TYPE_DBUS_TUBE,
        CONN_INTERFACE_ALIASING)

from defaults import UI_DIR
import sudoku


DBusGMainLoop(set_as_default=True)
logger = logging.getLogger("main.tubehandler")

def _get_contact_alias(conn, handle):
    return conn[CONN_INTERFACE_ALIASING].RequestAliases([handle])[0]

def _sanitize_service_name(service_name):
    return service_name.replace('.', '_')

def _build_bus_name(service_name):
    return "org.gnome.Empathy.DTubeHandler.%s" % \
            _sanitize_service_name(service_name)

def _build_object_path(bus_name):
    return '/' + bus_name.replace('.', '/')

class TubeHandler(dbus.service.Object):
    """Listen to coming tube

    tube_received_cb is called when the tube becomes open
    """

    def __init__(self, service_name, tube_received_cb):
        logger.debug("creating TubeHandler")
        self._tube_received_cb = tube_received_cb

        bus_name = _build_bus_name(service_name)
        obj_path = _build_object_path(bus_name)
        logger.debug("Bus name: %s" % bus_name)
        logger.debug("Object path: %s" % obj_path)

        bus = dbus.SessionBus()
        # this bus_name has to be retained, so make it a instance attribute
        self.busname = dbus.service.BusName(bus_name, bus)
        super(TubeHandler, self).__init__(bus, obj_path)

    @dbus.service.method(dbus_interface="org.gnome.Empathy.TubeHandler",
            in_signature='soouu', out_signature='')
    def HandleTube(self, bus_name, connection, channel, handle_type, handle):
        """Called when we are offered a tube"""
        logger.debug("Tube received on %s" % channel)
        self._bus_name = bus_name
        self._channel = channel
        self._peer_handle = handle
        self._conn = Connection(bus_name, connection,
                ready_handler=self._ready_handler)

    def _ready_handler(self, conn):
        self._tube_chan = Channel(self._bus_name, self._channel)
        self._tube_chan[CHANNEL_INTERFACE_TUBE].connect_to_signal(
                'TubeChannelStateChanged', self._tube_state_changed_cb)

        params = self._tube_chan[PROPERTIES_IFACE].Get(
                CHANNEL_INTERFACE_TUBE, 'Parameters')
        self._process_incoming_tube(params)

    def _tube_state_changed_cb(self, state):
        if state == TUBE_STATE_OPEN:
            logger.debug("Tube state changed ->open")
            bus = dbus.connection.Connection(self._address)
            self._tube_received_cb(bus, self._tube_chan,
                    self._puzzle, self._game_mode)
        else:
            logger.debug("Tube state changed ->%s" % state)

    def _process_incoming_tube(self, params):
        def reply_cb(dialog, response_id):
            if response_id == gtk.RESPONSE_OK:
                logger.debug("accept tube")
                self._address = self._tube_chan[CHANNEL_TYPE_DBUS_TUBE].Accept(
                        SOCKET_ACCESS_CONTROL_CREDENTIALS)
            else:
                logger.debug("deny tube offer")
                self._tube_chan[CHANNEL_INTERFACE].Close()
            dialog.destroy()

        try:
            puzzle = params["puzzle"]
            mode = params["game_mode"]
        except KeyError as err:
            logger.error("expected parameter missing %s" % err)
            return # do nothing

        self._puzzle = str(puzzle)
        self._game_mode = mode
        alias = _get_contact_alias(self._conn, self._peer_handle)
        logger.debug('got offer: %s' % self._puzzle)

        builder = gtk.Builder()
        builder.add_from_file(os.path.join(UI_DIR, 'network_game_offer.ui'))
        dialog = builder.get_object("dialog")
        dialog.connect("response", reply_cb)

        name_label = builder.get_object("name_label")
        name_label.set_text(_("%s wants to play with you") % alias)

        info = sudoku.SudokuRater(self._puzzle).difficulty()
        info_label = builder.get_object("info_label")
        info_label.set_text(_("Difficulty: %s ") % info.value_string())

        detail_label = builder.get_object("detail_label")
        detail_label.set_text(info.to_string())

        mode_str = _("Collaborative") if mode else _("Competitive")
        text = _("Game mode: %s") % mode_str
        mode_label = builder.get_object("mode_label")
        mode_label.set_text(text)

        dialog.run()
