# -*- coding: utf-8 -*-

import math
import random

import gobject
import gtk

import colors
import number_box
import model


TRACKER_COLORS = [
    # Use tango colors recommended here:
    # http://tango.freedesktop.org/Tango_Icon_Theme_Guidelines
    tuple([c / 255.0 for c in cols]) for cols in
    [(32, 74, 135), # Sky Blue 3
     (78, 154, 6), # Chameleon 3
     (206, 92, 0), # Orange 3
     (143, 89, 2), # Chocolate 3
     (92, 53, 102), # Plum 3
     (85, 87, 83), # Aluminium 5
     (196, 160, 0), # Butter 3
     ]
    ]

def _generate_new_color():
    random_color = TRACKER_COLORS[0]
    while random_color in TRACKER_COLORS:
        # If we have generated all possible colors, this will
        # enter an infinite loop
        random_color = (random.randint(0, 100)/100.0,
                        random.randint(0, 100)/100.0,
                        random.randint(0, 100)/100.0)
    return random_color


class _Tracker:
    def __init__(self, sudoku_view):
        self._trackers = {} # id => [coords]
        # FIXME what a bad name
        self.__trackers_tracking__ = {} # id => True/False
        self._tracker_by_coords = {} # coords => id
        self._sudoku_view = sudoku_view

    def add_to_tracker(self, tracker_id, x, y):
        self._trackers[tracker_id].append((x, y))
        self._tracker_by_coords[(x, y)] = tracker_id

    def remove_from_tracker(self, tracker_id, x, y):
        # TODO how to remove from self._trackers?
        self._tracker_by_coords[(x, y)] = None

    def get_active_tracker(self):
        for tracker_id in self.__trackers_tracking__:
            if self.__trackers_tracking__[tracker_id]:
                return tracker_id
        return None

    def get_tracker_by_coord(self, x, y):
        if (x, y) in self._tracker_by_coords:
            return self._tracker_by_coords[(x, y)]
        else:
            return None

    def create_tracker(self):
        identifier = 0
        while identifier in self._trackers:
            identifier += 1
        self._trackers[identifier] = []
        return identifier

    def toggle_tracker(self, identifier, flag):
        self.__trackers_tracking__[identifier] = flag

    def delete_by_tracker(self, identifier):
        """Delete all cells tracked by tracker ID identifier."""
        change = []
        # copy the list, or there will be infinite-loop
        entries = self._trackers[identifier]
        self._trackers[identifier] = []
        for (x, y) in entries:
            self._tracker_by_coords[(x, y)] = None
            change.append((x, y))
        self._sudoku_view.do_clear(change)

    def delete_except_for_tracker(self, identifier):
        for tracker_id in self._trackers:
            if not identifier == tracker_id:
                self.delete_by_tracker(tracker_id)

    def get_tracker_color(self, identifier):
        try:
            return TRACKER_COLORS[identifier]
        except IndexError:
            new_color = _generate_new_color()
            TRACKER_COLORS.append(new_color)
            return self.get_tracker_color(identifier)

    def get_trackers(self):
        return self._trackers

    def get_trackers_tracking(self):
        return self.__trackers_tracking__

    def reset(self):
        self._trackers = {}
        self.__trackers_tracking__ = {}
        self._tracker_by_coords = {}


def gtkcolor_to_rgb (color):
    return (color.red   / float(2**16),
            color.green / float(2**16),
            color.blue  / float(2**16))

SPACING_FACTOR = 40 # The size of a box compared (roughly) to the size
                    # of padding -- the larger this is, the smaller
                    # the spaces
SMALL_TO_BIG_FACTOR = 3.5 # The number of times wider than a small line a big line is.

class SudokuNumberGrid (gtk.AspectFrame):

    def __init__ (self, group_size):
        self.table = gtk.Table(rows = group_size, columns = group_size, homogeneous = True)
        self.group_size = group_size
        self.__entries__ = {}
        self._setup_table()

        gtk.AspectFrame.__init__(self, obey_child = False)
        self.set_shadow_type(gtk.SHADOW_NONE)
        self.eb = gtk.EventBox()
        self.eb.add(self.table)
        self.add(self.eb)
        self.connect('size-allocate', self._allocate_cb)
        self.show_all()

    def _setup_table(self):
        for x in range(self.group_size):
            for y in range(self.group_size):
                e = number_box.SudokuNumberBox(upper = self.group_size)
                e.x = x
                e.y = y
                self.table.attach(e, x, x+1, y, y+1)
                self.__entries__[(x, y)] = e
                e.connect('key-press-event', self._key_press_cb)

    def _allocate_cb (self, widget, rect):
        if rect.width > rect.height:
            side = rect.height
        else: side = rect.width
        # we want our small spacing to be 1/15th the size of a box
        spacing = float(side) / (self.group_size * SPACING_FACTOR)
        if spacing == 0:
            spacing = 1
        if hasattr(self, 'small_spacing') and spacing == self.small_spacing:
            return
        else:
            self._change_spacing(spacing)

    def _change_spacing (self, small_spacing):
        self.small_spacing = small_spacing
        self.big_spacing = int(small_spacing*SMALL_TO_BIG_FACTOR)
        self.table.set_row_spacings(int(small_spacing))
        self.table.set_col_spacings(int(small_spacing))
        box_side = int(math.sqrt(self.group_size))
        for n in range(1, box_side):
            self.table.set_row_spacing(box_side*n-1, self.big_spacing)
            self.table.set_col_spacing(box_side*n-1, self.big_spacing)
        self.table.set_border_width(self.big_spacing)

    def set_bg_color (self, color):
        if type(color) == str:
            try:
                color = gtk.gdk.color_parse(color)
            except (ValueError, TypeError): # TypeError prior to PyGTK 2.4
                print 'set_bg_color handed Bad color', color
                return
        self.eb.modify_bg(gtk.STATE_NORMAL, color)
        self.eb.modify_base(gtk.STATE_NORMAL, color)
        self.eb.modify_fg(gtk.STATE_NORMAL, color)
        self.table.modify_bg(gtk.STATE_NORMAL, color)
        self.table.modify_base(gtk.STATE_NORMAL, color)
        self.table.modify_fg(gtk.STATE_NORMAL, color)
        for e in self.__entries__.values():
            e.modify_bg(gtk.STATE_NORMAL, color)

    def get_value(self, x, y):
        return self.__entries__[(x, y)].get_value()

    def set_readonly_appearance(self, x, y, flag):
        self.__entries__[(x, y)].set_read_only(flag)

    def set_color(self, x, y, color):
        self.__entries__[(x, y)].set_color(color)

    def set_background_color(self, x, y, color):
        self.__entries__[(x, y)].set_background_color(color)

    def set_focus(self, x, y):
        self.table.set_focus_child(self.__entries__[x, y])

    def get_focus(self):
        return self.table.get_focus_child()

    def set_value(self, x, y, value):
        self.__entries__[(x, y)].set_value(value)

    def reset(self):
        pass

    def _key_press_cb(self, widget, event):
        key = gtk.gdk.keyval_name(event.keyval)
        dest_x, dest_y = self._go_around(widget.x, widget.y, key)
        if dest_x is not None:
            self.set_focus(dest_x, dest_y)

    def _go_around(self, x, y, direction):
        '''return the coordinate if we should go to the other side of the grid.
        Or else return None.'''
        limit_min, limit_max = 0, self.group_size - 1
        y_edge = [(limit_min, 'Up'), (limit_max, 'Down')]
        x_edge = [(limit_min, 'Left'), (limit_max, 'Right')]
        opposite = {limit_min: limit_max, limit_max: limit_min}

        if (y, direction) in y_edge:
            return x, opposite[y]
        elif (x, direction) in x_edge:
            return opposite[x], y
        else:
            return None, None


class SudokuView(SudokuNumberGrid):

    __gsignals__ = {
            "focus-changed": (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE,
                (int, int))
            }

    def __init__(self, group_size, controller=None):
        """User commands take action through controller. The GUI is updated by
        model. So in reality, it's like:
            user_input -> controller -> model -> view
        """
        SudokuNumberGrid.__init__(self, group_size)

        self._model = None
        self._do_highlight_cells = False
        self._notes_model = None

        self._controller = controller
        if controller:
            # if controller is not set, we don't react to any user interaction
            for e in self.__entries__.values():
                e.show()
                e.connect('number-changed', self._number_changed_cb)
                e.connect('notes-changed', self._notes_changed_cb)
                e.connect('focus-in-event', self._focus_callback)

        self._tracker = _Tracker(self)

    def reset(self):
        for x in range(self.group_size):
            for y in range(self.group_size):
                self.set_value(x, y, 0)
                self.set_readonly_appearance(x, y, False)
                self.set_color(x, y, None)
        self._tracker.reset()
        self._reset_notes()

    def connect_to_model(self, sudoku_model):
        """Listening update from sudoku_model
        """
        if not sudoku_model:
            return
        self._model = sudoku_model
        self._model.add_observer(self)
        for x in range(self.group_size):
            for y in range(self.group_size):
                val = self._model.get_value(x, y)
                if val:
                    self.set_value(x, y, val)
                    if self._model.get_virgin_value(x, y):
                        self.set_readonly_appearance(x, y, True)

    def connect_to_notes_model(self, notes_model):
        self._notes_model = notes_model
        self._notes_model.add_observer(self)

    def update(self, values):
        """Update the GUI, called by a Model

        values should be a list of NumberBoxModel"""
        for box in values:
            x, y, value, conflict = box.x, box.y, box.value, box.conflict
            is_new_value = value is not None
            if is_new_value:
                self.set_value(x, y, value)
            self.set_color(x, y,
                    self._refresh_color(x, y, conflict, is_new_value))

    def set_value(self, x, y, value):
        if not value:
            # setting to empty should also delete it from trackers
            tid = self._tracker.get_tracker_by_coord(x, y)
            if tid is not None:
                self._tracker.remove_from_tracker(tid, x, y)
        SudokuNumberGrid.set_value(self, x, y, value)

    def _refresh_color(self, x, y, conflict, is_new_value):
        """Set the color according to several rules, conflict/tracker, etc
        """
        # conflict
        if conflict:
            return (1.0, 0, 0)

        # there is active tracker
        tid = self._tracker.get_active_tracker()
        if tid is not None and is_new_value: # old values won't go into tracker
            self._tracker.add_to_tracker(tid, x, y)
            return self._tracker.get_tracker_color(tid)

        # the entry was in a tracker; should resume it if its color has been
        # overridden by conflict
        tid = self._tracker.get_tracker_by_coord(x, y)
        if tid is not None:
            return self._tracker.get_tracker_color(tid)

        # default black color
        return None

    def update_notes(self, notes_list):
        for notes in notes_list:
            x, y, top_note, bottom_note = notes
            self.__entries__[(x, y)].set_notes((top_note, bottom_note))

    def _reset_notes(self):
        for e in self.__entries__.values():
            e.set_notes(("", ""))

    def _number_changed_cb(self, widget, new_number):
        x, y = widget.x, widget.y
        # Use the displayed value, instead of that from the grid;
        # because conflicting value doesn't get into the grid
        old = self.get_value(x, y)
        cmd = model.ChangeValueCmd(self._model, x, y, old, new_number)
        self._controller.execute_command(cmd)

    def _get_changed_list(self):
        """Return all the boxes that are changed (compared to virgin puzzle)
        """

        res = []
        for x in range(self.group_size):
            for y in range(self.group_size):
                if not self._model.get_virgin_value(x, y):
                    # conflicting values are not being tracked, so we have to
                    # set_value on all the entries, instead of only filled ones
                    res.append((x, y))
        return res

    def do_clear(self, change_list=None):
        res = []
        if not change_list:
            change_list = self._get_changed_list()
        for x, y in change_list:
            old = self.get_value(x, y)
            res.append((x, y, old, 0))
        cmd = model.MultiChangeValueCmd(self._model, res)
        self._controller.execute_command(cmd)

    def _notes_changed_cb(self, widget, top_note, bottom_note):
        self._notes_model.set_notes(widget.x, widget.y, top_note, bottom_note)

    def _focus_callback(self, widget, event):
        self.emit("focus-changed", widget.x, widget.y)
        if self._do_highlight_cells:
            self._highlight_cells()

    def toggle_highlight (self, val):
        self._do_highlight_cells = val
        if val:
            self._highlight_cells()
        else:
            self.unhighlight_cells()

    def unhighlight_cells(self):
        for e in self.__entries__.values():
            e.set_background_color(None)

    def _highlight_cells(self):
        focused = self.get_focus()
        if not focused: # no need to do anything if no box has focus
            return
        self.unhighlight_cells() # first clear all the cells
        if not hasattr(self, 'box_color'):
            self._get_highlight_colors()
        my_x, my_y = focused.x, focused.y

        grid = self._model.grid
        for x, y in grid.col_coords[my_x]:
            if (x, y) != (my_x, my_y):
                self.__entries__[(x, y)].set_background_color(self.col_color)
        for x, y in grid.row_coords[my_y]:
            if (x, y) != (my_x, my_y):
                self.__entries__[(x, y)].set_background_color(self.row_color)
        for x, y in grid.box_coords[grid.box_by_coords[(my_x, my_y)]]:
            if (x, y) != (my_x, my_y):
                e = self.__entries__[(x, y)]
                if x == my_x:
                    e.set_background_color(self.box_and_col_color)
                elif y == my_y:
                    e.set_background_color(self.box_and_row_color)
                else:
                    e.set_background_color(self.box_color)

    def _get_highlight_colors(self):
        entry = self.__entries__.values()[0]
        default_color = gtkcolor_to_rgb(entry.style.bg[gtk.STATE_SELECTED])
        hsv = colors.rgb_to_hsv(*default_color)
        box_s = hsv[1]
        box_v = hsv[2]
        if box_v < 0.5:
            box_v = box_v * 2
        if box_s > 0.75:
            box_s = box_s * 0.5
        else:
            box_s = box_s * 1.5
            if box_s > 1:
                box_s = 1.0
        self.box_color = colors.hsv_to_rgb(hsv[0], box_s, box_v)
        self.box_and_row_color = colors.rotate_hue_rgb(*self.box_color, **{'rotate_by': 0.33 / 2})
        self.row_color = colors.rotate_hue_rgb(*self.box_color, **{'rotate_by': 0.33})
        self.col_color = colors.rotate_hue_rgb(*self.box_color, **{'rotate_by': 0.66})
        self.box_and_col_color = colors.rotate_hue_rgb(*self.box_color, **{'rotate_by': 1.0 - (0.33 / 2)})

    def get_tracker_controller(self):
        return self._tracker

    def get_trackers(self):
        # We have to save not only the coordinates, but also the value.
        # Maybe the values should be also tracked by _Tracker?
        result = {}
        trackers = self._tracker.get_trackers()
        for tid in trackers:
            result[tid] = [(x, y, self.get_value(x, y)) for (x, y) in
                trackers[tid]]
        return result

    def get_trackers_tracking(self):
        return self._tracker.get_trackers_tracking()

class SudokuSideView(SudokuNumberGrid):
    def __init__(self, group_size):
        self._focus = None
        self._focus_bg = None

        SudokuNumberGrid.__init__(self, group_size)
        for x in range(self.group_size):
            for y in range(self.group_size):
                self.set_readonly_appearance(x, y, True)

    def reset(self):
        for x in range(self.group_size):
            for y in range(self.group_size):
                self.set_value(x, y, 0)
                self.set_background_color(x, y, None)
        self._focus = None
        self._focus_bg = None

    def set_invisible(self, x, y):
        self.set_background_color(x, y, (0, 0, 0))
        if self._focus and (x, y) == (self._focus.x, self._focus.y):
            self._focus_bg = (0, 0, 0)

    def set_empty(self, x, y):
        self.set_value(x, y, 0)
        self.set_background_color(x, y, None)

    def set_focus(self, x, y):
        # restore previously focused entry
        if self._focus:
            self.set_background_color(self._focus.x, self._focus.y,
                    self._focus_bg)
        # highlight new focused entry
        self._focus = self.__entries__[(x, y)]
        self._focus_bg = self._focus.get_background_color()
        self.set_background_color(x, y, (0, 1, 0))
        # this _focus/_focus_bg is not at all elegant, but for now it does what
        # I need, should come up with a better solution
